// Aspell Data Objects Implementations 
// Copyright 1998 Under the terms of the LGPL

//
// Aspell's main word list data is stored in 4 large blocks of memory
//
// * The Word Hash Table
// * The Word List
// * The Soundslike Hash Table
// * The Soundslike List
//
// 1a) The Word Hash Table
// This consists of an open address hash table which contains pointers
// to the actual words in the word list
//
// 1b) The Word List
// This consists of the actual word list and is layed out as follows:
//   <Word1><null char><Word2><null char>...
//
// 2a) The Soundslike Hash Table
// This consists of an open address hash table which contains pointers
// to a soundslike object.
//
// 2b) The Soundslike Object
// The soundslike object is layed out as follow:
//  <Word1 pointer><Word2 pointer>...<Num of Words><Soundslike><null char>
//  <const char *><const char *>...<unsigned short int><char[]><char>
// The pointer to the object points to the beginning of the Soundslike string
// The Word pointers consists of the the words which have the same 
//   soundslike pattern
//
// 2c) The Soundslike List
// This consists of Soundslike Objects back to back:
//  <Soundslike object 1><Soundslike object 2> ...
// There is no delimiter between the objects
//
//
//                          Format of the *.wrd files
//
// (This part is in ascii format)
// <"master_wl"><ws><lang name><ws><# words><ws>
//     <hash size><ws><size of list block><\n>
// (The rest is in binary format>
// <Wordlist>
// <Word Hash Table>
//
// The word hash table is a vector of unsigned its which contains an offset
// of where they can be found in the word list.
//
//                          Format of the *.sl files
//
// (This part is in ascii format)
// <"master_wl"><ws><lang name><ws><# words><ws>
//     <hash size><ws><size of list block><\n>
// (The rest is in binary format>
// <Soundslike object list>
// <Soundslike Hash Table>
//
// Soundslike oject is laid out as follows:
//   <Num of Words><Word 1 offset>...<Soundslike><\0>
//   <unsigned short int><unsigned int>...<char[]><char>
// And like the .wrd file the hash table contains offsets not pointers.
//

#include <string>
#include <fstream.h>
#include <strstream.h>

#include <time.h>
#include <sys/stat.h>

#include "error.hh"
#include "data.hh"
#include "real_data.hh"
#include "language.hh"
#include "hash-t.hh"
#include "file.hh"

void SC_ReadOnly_WordList::load(const string & file_name,
				const AspellConfig * config) 
{
  ifstream WORDS;
  open_file(WORDS, file_name, ios::in | ios::bin);

  string check_word;
  WORDS >> check_word;
  if (check_word != "master_wl")
    throw ae_bad_format(file_name, 0);

  WORDS >> check_word;

  try {
    parent_->set_check_lang(check_word,config);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, 0);
  }

  int count, bucket, size;
  WORDS >> count >> bucket >> size;
  WORDS.ignore();
  master_ = new char[size];

  WORDS.read(master_, size);

  word_list.table().resize(bucket);
  WordList::table_type::iterator i   = word_list.table().begin();

  unsigned int pos;
  while (!WORDS.eof()) {
    WORDS.read(&pos, sizeof(unsigned int));
    *i = master_ + pos;
    ++i;
  }

  word_list.set_size(count);
}

void SC_ReadOnly_Soundslikes::load(const string &file_name, const char *d,
				   const AspellConfig * config)
{
  ifstream SOUNDS;
  open_file(SOUNDS, file_name, ios::in | ios::bin);
  
  string check_word;
  SOUNDS >> check_word;
  if (check_word != "master_ml") 
    throw ae_bad_format(file_name, 0);

  SOUNDS >> check_word;

  try {
    parent_->set_check_lang(check_word,config);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, 0);
  }

  int count, bucket, size;
  SOUNDS >> count >> bucket >> size;
  main_data = new char[size];
  empty_wordlist = main_data + sizeof(unsigned short int);
  char *data = main_data;
  
  SOUNDS.ignore();

  *reinterpret_cast<unsigned short int *>(data) = 0;
  *(data + sizeof(unsigned short int *)) = '\0';
  data = main_data + sizeof(const char *);
  
  unsigned short int list_count;
  unsigned int       pos;
  unsigned int       m;
  while (SOUNDS.read(reinterpret_cast<void*>(&list_count),sizeof(unsigned short int)),
	 list_count) {
    for(unsigned short int i=0; i != list_count; ++i) {
      SOUNDS.read(&pos, sizeof(unsigned int));
      *reinterpret_cast<const char **>(data) = d + pos;
      data += sizeof(const char *);
    }
    *reinterpret_cast<unsigned short int *>(data) = list_count;
    data += sizeof(unsigned short int);
    while ((*data = SOUNDS.get()) != '\0') ++data;
    ++data;
    m = (data - main_data) % sizeof(const char *);
    if (m != 0)
      data += sizeof(const char *) - m;
  }

  lookup_table.table().resize(bucket);

  LookupTable::table_type::iterator i   = lookup_table.table().begin();

  while (!SOUNDS.eof()) {
    SOUNDS.read(&pos, sizeof(unsigned int));
    *i = main_data + pos;
    ++i;
  }

  lookup_table.set_size(count);
}

const SC_ReadOnly_Soundslikes::WordList SC_ReadOnly_Soundslikes
::words_w_soundslike(const char *word) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end() || !(*i).size()) { 
    //                                  ^^ FIXME
    // There should not be a need for the size check as there should never be
    // any empty lists in the lookup_table.
    return empty_wordlist;
  } else {                                       
    return *i;
  }
}

const SC_ReadOnly_Soundslikes::WordList SC_ReadOnly_Soundslikes
::words_w_soundslike(const char *word, SC_Sound &M) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end() || !(*i).size()) {
    //                                  ^^ FIXME
    // There should not be a need for the size check as there should never be
    // any empty lists in the lookup_table.
    return empty_wordlist;
  } else {
    M = (*i).key();
    return *i;
  }
}

void SC_Writable_Soundslikes::add(const char* w, const string &s) {
  lookup_table[s.c_str()].push_back(w);
}

void SC_Writable_Soundslikes::erase(const char* w, const string &s) {
  WordList& item = lookup_table.find(s.c_str())->second;
  WordList::iterator i = item.begin();
  WordList::iterator end = item.end();
  while(i != end && *i != w) ++i;
  if (i == end) return;
  item.erase(i);
  if (!item.size()) {
    lookup_table.erase(s.c_str());
  }
}

const SC_Writable_Soundslikes::WordList * SC_Writable_Soundslikes
::words_w_soundslike(const char *word) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end()) {
    return NULL;
  } else {
    return &(i->second);
  }
}

const SC_Writable_Soundslikes::WordList * SC_Writable_Soundslikes
::words_w_soundslike(const char *word, SC_Sound &M) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end()) {
    return NULL;
  } else {
    M = i->first.c_str();
    return &(i->second);
  }
}

//
//
//

struct ReadOnlyWL::Internal {
  string wl_file_name;
  string sl_file_name;
  SC_ReadOnly_WordList    words;
  SC_ReadOnly_Soundslikes soundslikes;
  Internal(ReadOnlyWL * w) : words(w), soundslikes(w) {}
};

const char * ReadOnlyWL::file_name() const {
  //FIXME:  This should return the base (file_name - ext) for the file name
  return internal->wl_file_name.c_str();
}

ReadOnlyWL::const_iterator ReadOnlyWL::begin() const {
  return make_virtual_forward_iteratorr<const char *>(internal->words.begin());
}

ReadOnlyWL::const_iterator ReadOnlyWL::end() const {
  return make_virtual_forward_iteratorr<const char *>(internal->words.end());
}

ReadOnlyWL::size_type ReadOnlyWL::size() const {
  return internal->words.size();
}

bool ReadOnlyWL::empty() const {
  return internal->words.empty();
}

ReadOnlyWL::ReadOnlyWL() {
  internal = new Internal(this);
}

ReadOnlyWL::~ReadOnlyWL() {
  delete internal;
}

void ReadOnlyWL::load(const string & b,
		      const AspellConfig * config)
{
  internal->wl_file_name = b + ".wrd";
  internal->sl_file_name = b + ".sl";
  internal->words.load(internal->wl_file_name, config);
  internal->soundslikes.load(internal->sl_file_name, 
			     internal->words.get_word_block(),
			     config);
}

bool ReadOnlyWL::operator [] (const char *word) const {
  return internal->words[word];
}

bool ReadOnlyWL::operator [] (const string &word) const {
  return internal->words[word];
}

const ReadOnlyWL::WordList ReadOnlyWL::words_w_soundslike(const char *word) const {
  const SC_ReadOnly_Soundslikes::WordList 
    temp = internal->soundslikes.words_w_soundslike(word);
  return make_virtual_const_begin_end_sizer<WordList>(temp);
}

//
//
//

static time_t modification_date(const string &file) {
  struct stat file_stat;
  if (stat(file.c_str(), &file_stat) == 0)
    return file_stat.st_mtime;
  else 
    return 0;
}

//
//
//

struct WritableWL::Internal {
  string cur_file_name;
  time_t cur_file_date;
  WritableWL * real_wl_;
  SC_Writable_WordList    words;
  SC_Writable_Soundslikes soundslikes;
  static const SC_Writable_Soundslikes::WordList empty_wordlist;
  Internal(WritableWL *w) : real_wl_(w), words(w), soundslikes(w) {}
  void save_as(const string &);
  void merge(const string &, const AspellConfig * config = 0);
};

const SC_Writable_Soundslikes::WordList WritableWL::Internal::empty_wordlist;

struct SC_WWL_deref_via_cstr {
  const char * operator() (SC_Writable_WordList::const_iterator i) const
    {return i->c_str();}
};

WritableWL::const_iterator WritableWL::begin() const {
  return make_virtual_forward_iteratorr<const char *>
    (alt_dereference_iterator<SC_Writable_WordList::const_iterator,
                              SC_WWL_deref_via_cstr,
                              const char *, const char *, const char **>
     (internal->words.begin())
    );
}

WritableWL::const_iterator WritableWL::end() const {
  return make_virtual_forward_iteratorr<const char *>
    (alt_dereference_iterator<SC_Writable_WordList::const_iterator,
                              SC_WWL_deref_via_cstr,
                              const char *, const char *, const char **>
     (internal->words.end())
    );
  //return make_virtual_forward_iteratorr<const char *>(internal->words.end());
}

WritableWL::size_type WritableWL::size() const {
  return internal->words.size();
}

bool WritableWL::empty() const {
  return internal->words.empty();
}

void WritableWL::Internal::merge(const string &file_name, 
				 const AspellConfig * config)
{
  ifstream WORDS;
  open_file(WORDS, file_name);

  int c;
  string word, sound;
  const char *word2;
  
  WORDS >> word;
  if (word != "personal_wl")
    throw ae_bad_format(file_name, real_wl_);

  WORDS >> word;

  try {
    real_wl_->set_check_lang(word, config);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, real_wl_);
  }

  WORDS >> c; // not used at the moment
  
  for (;;) {
    WORDS >> word >> sound;
    if (!WORDS) break;
    word2 = words.add(word);
    soundslikes.add(word2, sound);
  }

}

void WritableWL::Internal::save_as(const string &file_name) {
  ofstream WORDS((file_name).c_str());

  if (WORDS.rdstate() & ios::badbit) 
    throw ae_cant_write(file_name, real_wl_);

  WORDS << "personal_wl" << ' ' << real_wl_->lang_name() <<  endl;
  WORDS << words.size() << endl;

  SC_Writable_Soundslikes::const_iterator i = soundslikes.begin();
  SC_Writable_Soundslikes::const_iterator e = soundslikes.end();

  SC_Writable_Soundslikes::WordList::const_iterator j;
  
  for (;i != e; ++i) {
    for (j = i->second.begin(); j != i->second.end(); ++j) {
      WORDS << *j << ' ' << i->first << endl;
    }
  }

  WORDS.close();
  cur_file_date = modification_date(file_name);
  
}

WritableWL::WritableWL() {
  internal = new Internal(this);
}

WritableWL::~WritableWL() {
  delete internal;
}

void WritableWL::load(const string & file_name, 
		      const AspellConfig * config)
{
  internal->cur_file_name = file_name;
  internal->cur_file_date = modification_date(file_name);
  internal->merge(file_name, config);
}

void WritableWL::merge(const string & file_name) {
  internal->merge(file_name);
}

void WritableWL::update() {
  internal->cur_file_date = modification_date(internal->cur_file_name);
  internal->merge(internal->cur_file_name);
}

void WritableWL::save_as(const string &b) {
  return internal->save_as(b);
}

void WritableWL::save() {
  time_t new_date = modification_date(internal->cur_file_name);
  if (new_date > internal->cur_file_date)
    update();
  internal->save_as(internal->cur_file_name);
  internal->cur_file_date = modification_date(internal->cur_file_name);
}

const char * WritableWL::file_name() const {
  return internal->cur_file_name.c_str();
}

void WritableWL::add(const string &w) {
  if (!operator[](w))
    internal->soundslikes.add(internal->words.add(w),
			      lang()->to_soundslike(w));
}

void WritableWL::add(const string &w, const string &s) {
  if (!operator[](w))
    internal->soundslikes.add(internal->words.add(w),s);
}

void WritableWL::clear() {
  internal->words.clear(); 
  internal->soundslikes.clear();
}

bool WritableWL::operator [] (const char *word) const {
  return internal->words[word];
}

bool WritableWL::operator [] (const string &word) const {
  return internal->words[word];
}

const WritableWL::WordList WritableWL::words_w_soundslike(const char *word) const {
  const SC_Writable_Soundslikes::WordList * 
    temp = internal->soundslikes.words_w_soundslike(word);
  if (temp) {
    return make_virtual_const_begin_end_sizer<WordList>(*temp);
  } else {
    return make_virtual_const_begin_end_sizer<WordList>(Internal::empty_wordlist);
  }
}

//
// FIXME:  Fix file nameing issue
// so that aspell create repl < tmp will WORK!!!!
// change class file_name to cur_file_name
// and same with cur_file_date
//

bool WritableReplList::RealReplacements::exists(const string &word) {
  iterator i = begin();
  iterator e = end();
  while (i != e) {
    if (*i == word) return true;
    ++i;
  }
  return false;
}

bool WritableReplList::RealReplacements::add(const string &word) {
  if (exists(word)) return false;
  info.push_back(word);
  return true;
}

bool WritableReplList::RealReplacements::erase(const string &word) {
  vector<string>::iterator i = info.begin() + 1;
  vector<string>::iterator e = info.end();
  while (i != e) {
    if (*i == word) {
      info.erase(i);
      return true;
    }
    ++i;
  }
  return false;
}

// REAL lookup table:
// hash_map<string, vector<RealReplacements, ..> lookup_table
//   RealRepl.. is a custom struct where begin() and end() are
//     iterators for the repl. list for a misspelled word
//   The _elements_ of the hash_map are a vector of RealRepl.
//      where the mispelled_word all have a soundslike eqilvent
//      to _key_.

struct WritableReplList::Internal {
  typedef vector<RealReplacements> ReplList;
  typedef hash_map<string, ReplList, hash_string<string> >
           LookupTable;

  struct deref_via_cstr {
    const char * operator() (RealReplacements::const_iterator i) const
      {return i->c_str();}
  };
  struct deref_to_repl {
    Replacements operator() (ReplList::const_iterator i) const {
      return Replacements(i->misspelled_word().c_str(),
			  repl_vir_itr(i->begin()),
			  repl_vir_itr(i->end()),
			  i->size());
    }
  };
  typedef alt_deref_make_virtual_forward_iterator
                 < RealReplacements::const_iterator, deref_via_cstr, 
                   const char *, const char *, const char**>
          repl_vir_itr;
  typedef alt_deref_make_virtual_forward_iterator
                < ReplList::const_iterator, deref_to_repl, 
	          Replacements, Replacements, Replacements* > 
          repl_list_vir_itr;

  struct get_wordlist {
    WordList operator() (LookupTable::const_iterator i) const {
      return WordList(repl_list_vir_itr(i->second.begin()),
		      repl_list_vir_itr(i->second.end()),
		      i->second.size());
    }
  };

  // this class is needed because the using the generic 
  // alt_derefrence_iterator caused problem on metalab (solaris)
  struct lookup_table_alt_deref_itr {  
    typedef LookupTable::const_iterator Itr;

    typedef iterator_traits<Itr>::iterator_category iterator_category;
    typedef iterator_traits<Itr>::difference_type   difference_type;
    typedef WordList                                   value_type;
    typedef WordList                                   reference;
    typedef WordList *                                 pointer;

  private:
    Itr          itr;
    get_wordlist deref;
  public:
    reference operator*() const {return deref(itr);}
    lookup_table_alt_deref_itr& operator++() {++itr; return *this;}
    lookup_table_alt_deref_itr operator++(int) 
      {lookup_table_alt_deref_itr temp=*this; ++itr; return temp;}
    bool is_equal(lookup_table_alt_deref_itr other) const 
      {return itr == other.itr;}
    bool operator== (lookup_table_alt_deref_itr other) const 
      {return is_equal(other);}
    bool operator!= (lookup_table_alt_deref_itr other) const
      {return !is_equal(other);}
    lookup_table_alt_deref_itr(const LookupTable::iterator & i): itr(i) {}
  };

  static const ReplList empty_wordlist0;
  static const WordList empty_wordlist;

  WritableReplList * real_wl_;
  string cur_file_name;
  time_t cur_file_date;
  LookupTable * lookup_table;

  Internal(WritableReplList * w) : real_wl_(w) 
    {lookup_table = new LookupTable();}
  ~Internal() {delete lookup_table;}
  void save_as(const string &);
  void clear();
  void load(const string &, const AspellConfig *);
  void merge(const string &);
};

#define Inter WritableReplList::Internal

const Inter::ReplList            Inter::empty_wordlist0;
const WritableReplList::WordList Inter::empty_wordlist
  (repl_list_vir_itr(empty_wordlist0.begin()),
   repl_list_vir_itr(empty_wordlist0.end()),
   empty_wordlist0.size());

#undef Inter


WritableReplList::WritableReplList() {
  internal = new Internal(this);
}

WritableReplList::~WritableReplList() {
  delete internal;
}

WritableReplList::const_iterator WritableReplList::begin() const {
  Internal::lookup_table_alt_deref_itr 
    b = Internal::lookup_table_alt_deref_itr(internal->lookup_table->begin()),
    e = Internal::lookup_table_alt_deref_itr(internal->lookup_table->end());
  return
    make_virtual_forward_iteratorr<Replacements>(
      multi_forward_iterator<Internal::lookup_table_alt_deref_itr>
                                                            (b,e,(*b).begin())
    );
}


WritableReplList::const_iterator WritableReplList::end()   const {
  Internal::lookup_table_alt_deref_itr 
    e = Internal::lookup_table_alt_deref_itr(internal->lookup_table->end());
  return
    make_virtual_forward_iteratorr<Replacements>(
      multi_forward_iterator<Internal::lookup_table_alt_deref_itr> (e,e)
    );
}

WritableReplList::size_type WritableReplList::size() const {return 0;}

bool WritableReplList::empty() const {
  return internal->lookup_table->empty();
}

void WritableReplList::add(const string &mis, const string &cor) {
  add(mis, cor, lang()->to_soundslike(mis));
}

void WritableReplList
::add(const string &mis, const string &cor, const string &s) {
  Internal::LookupTable::iterator i = internal->lookup_table->find(s);
  if (i == internal->lookup_table->end())
    i = internal->lookup_table->insert
          (Internal::LookupTable::value_type(s, Internal::ReplList())).first;
  
  Internal::ReplList::iterator    j = i->second.begin();
  Internal::ReplList::iterator    e = i->second.end();
  for (; j != e; ++j) {
    if ((*j).misspelled_word() == mis) {
      (*j).add(cor);
      return;
    }
  }
  i->second.push_back(RealReplacements(mis,cor));
}

const WritableReplList::WordList 
WritableReplList::words_w_soundslike(const char *word) const {
  Internal::LookupTable::const_iterator i = internal->lookup_table->find(word);
  if (i == internal->lookup_table->end()) {
    return Internal::empty_wordlist;
  } else {
    return WordList(Internal::repl_list_vir_itr(i->second.begin()),
		    Internal::repl_list_vir_itr(i->second.end()),
		    i->second.size());
  }
}

void WritableReplList::Internal::save_as(const string & file_name) {

  ofstream WORDS((file_name).c_str());

  if (WORDS.rdstate() & ios::badbit) 
    throw ae_cant_write(file_name, real_wl_);

  WORDS << "personal_repl" << ' ' << real_wl_->lang_name() <<  endl;
  WORDS << lookup_table->size() << ' ' << lookup_table->bucket_count() << endl;
  
  LookupTable::iterator i = lookup_table->begin();
  LookupTable::iterator e = lookup_table->end();
  
  for (;i != e; ++i) {
    WORDS << i->first << ' ' << i->second.size() << endl;
    for (ReplList::iterator j = i->second.begin(); j != i->second.end(); ++j) {
      WORDS << "  " << (*j).misspelled_word() << ' ' << (*j).size() << ' ';
      for (RealReplacements::iterator k = j->begin(); k != j->end(); ++k) {
	WORDS << *k << ',';
      }
      WORDS << endl;
    }
  }

  WORDS.close();
  
  cur_file_date = modification_date(file_name);
}

void WritableReplList::save_as(const string &b) {
  internal->save_as(b);
}

void WritableReplList::save() {
  time_t new_date = modification_date(internal->cur_file_name);
  if (new_date > internal->cur_file_date)
    update();
  internal->save_as(internal->cur_file_name);
  internal->cur_file_date = modification_date(internal->cur_file_name);
}

void WritableReplList::Internal::clear() {
  delete lookup_table;
  lookup_table = new LookupTable();
}

void WritableReplList::Internal::load(const string & file_name, 
				      const AspellConfig * config)
{
  cur_file_name = file_name;
  cur_file_date = modification_date(file_name);

  ifstream WORDS;
  open_file(WORDS, file_name);

  int c;
  string word, mis, sound, repl;
  int num_words, num_repls;

  WORDS >> word;
  if (word != "personal_repl")
    throw ae_bad_format(file_name, real_wl_);

  WORDS >> word;

  try {
    real_wl_->set_check_lang(word, config);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, real_wl_);
  }

  int num_soundslikes;
  WORDS >> num_soundslikes;
  WORDS >> c;  // not used at the moment

  int h,i,j;
  for (h=0; h != num_soundslikes; ++h) {
    WORDS >> sound >> num_words;
    ReplList &wl = lookup_table->insert(
      LookupTable::value_type(sound,ReplList())).first->second;
    wl.reserve(num_words);
    for (i = 0; i != num_words; ++i) {
      WORDS >> mis >> num_repls;
      WORDS.ignore(); // ignore space
      wl.push_back(RealReplacements(mis,num_repls));
      RealReplacements &rp = *(wl.end()-1);
      for (j = 0; j != num_repls; ++j) {
	getline(WORDS, repl, ',');
	rp.add_nocheck(repl);
      }
    }
  }
}

void WritableReplList::Internal::merge(const string &file_name) {

  ifstream WORDS;
  open_file(WORDS, file_name);

  int c;
  string word, mis, sound, repl;
  int num_words, num_repls;

  WORDS >> word;
  if (word != "personal_repl")
    throw ae_bad_format(file_name, real_wl_);

  WORDS >> word;

  try {
    real_wl_->check_lang(word);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, real_wl_);
  }

  int num_soundslikes;
  WORDS >> num_soundslikes;
  WORDS >> c;  // not used at the moment

  int h,i,j;
  ReplList::iterator k;
  bool found;
  RealReplacements * rp;
  for (h=0; h != num_soundslikes; ++h) {
    WORDS >> sound >> num_words;
    ReplList &wl = lookup_table->insert(
      LookupTable::value_type(sound,ReplList())).first->second;
    for (i = 0; i != num_words; ++i) {
      WORDS >> mis >> num_repls;
      WORDS.ignore(); // ignore space
      found = false;
      for (k = wl.begin(); k != wl.end(); ++k) {
	if (mis == (*k).misspelled_word()) {
	  found = true;
	  break;
	}
      }
      if (found) {
	rp = &(*k);
      } else {
	wl.push_back(RealReplacements(mis,num_repls));
        rp = &(*(wl.end()-1));
      }
      for (j = 0; j != num_repls; ++j) {
	getline(WORDS,repl,';');
	rp->add(repl);
      }
    }
  }
}

void WritableReplList::clear() {
  internal->clear();
}

void WritableReplList::load(const string & file_name,
			    const AspellConfig * config)
{
  internal->load(file_name, config);
}

void WritableReplList::merge(const string &b) {
  internal->merge(b);
}

void WritableReplList::update() {
  internal->cur_file_date = modification_date(internal->cur_file_name);
  internal->merge(internal->cur_file_name);
}

const char * WritableReplList::file_name() {
  return internal->cur_file_name.c_str();
}


