
#include "aspell.hh"
#include "data.hh"
#include "language.hh"
#include "config.hh"

#include "copy_ptr-t.hh"

template copy_ptr<const SC_Language>;
template copy_ptr<AspellConfig>;

//
// data_access functions
//

aspell::aspell(const StringMap & override) {
  setup(override);
}

aspell::aspell(const string & override) {
  StringMap temp;
  temp.read_in_string(override);
  setup(temp);
}

void aspell::setup(const StringMap & override)
{
  config_ = new AspellConfig(override);
  ignore_repl = config_->have("ignore_repl");
  try {
    {
      StandardWL_Base * temp = new ReadOnlyWL;
      temp->load(config_->lookup("master-path"), config_.ptr);
      change_id(attach(temp), main_id)->own(true);
    } {
      StandardWL_Base * temp = new WritableWL;
      change_id(attach(temp), personal_id)->own(true);
      try {temp->load((*config_)["personal-path"],config_.ptr);} 
      catch (ae_cant_read&) {}
    } {
      change_id(attach(new WritableWL), session_id)->own(true);
    } {
      ReplacementsWL_Base * temp = new WritableReplList;
      change_id(attach(temp), per_repl_id)->own(true);
      try {temp->load((*config_)["repl-path"],config_.ptr);} 
      catch (ae_cant_read&) {}
    }
  } catch (...) {
    destroy();
    throw;
  }
}

aspell::aspell(): ignore_repl(true)
{}

aspell::~aspell() {
  destroy();
}

const char * aspell::lang_name() const {
  return lang_->name();
}

void aspell::add_personal(const string &word) {
  WordLists::iterator i = locate_wl(personal_id);
  if (i != word_lists_.end())
    static_cast<WritableWL_Base *>(i->wl())->add(word);
}

void aspell::add_session(const string &word) {
  WordLists::iterator i = locate_wl(session_id);
  if (i != word_lists_.end())
    static_cast<WritableWL_Base *>(i->wl())->add(word);
}

void aspell::clear_session() {
  WordLists::iterator i = locate_wl(session_id);
  if (i != word_lists_.end())
    static_cast<WritableWL_Base *>(i->wl())->clear();
}

bool aspell::ignore_repls() {
  return ignore_repl;
}

bool aspell::ignore_repls(bool r) {
  bool temp = ignore_repl;
  if (r) {
    ignore_repl = true;
    config_->replace("ignore_repl","true");
  } else {
    ignore_repl = false;
    config_->replace("ignore_repl","false");
  }
  return temp;
}

void aspell::store_repl(const string& mis, const string &cor, bool memory) {
  WordLists::iterator i = locate_wl(per_repl_id);
  if (!ignore_repl && i != word_lists_.end()) {
    string::size_type pos;
    Suggestions sugs = suggest(mis);
    if (pos = cor.find(' '), 
	pos == string::npos 
	? check(cor)
	: (check(cor.substr(0,pos)) && check(cor.substr(pos+1))) )
      { // cor is a correct spelling
	
	if (sugs.empty() || cor != sugs[0])
	  static_cast<ReplacementsWL_Base *>(i->wl())->add(mis, cor);

	if (memory && prev_cor_repl_ == mis) 
	  store_repl(prev_mis_repl_, cor, false);
	
      } else { // cor is not a correct spelling
	
	if (memory) {
	  if (prev_cor_repl_ != mis)
	    prev_mis_repl_ = mis;
	  prev_cor_repl_ = cor;
	}
      }
  }
}

//
// simple functions
//

inline bool aspell::check_notrim(const string &word) const {
  if (check_raw(word)) return true;
  else return check_raw(lang_->to_lower(word));
}

inline bool aspell::check_notrim(const char *word) const {
  if (check_raw(word)) return true; 
  else return check_raw(lang_->to_lower(word));
}

bool aspell::check(const string &word) const {
  if (check_notrim(word)) return true;
  return lang_->trim_n_try(*this, word);
}

bool aspell::check(const char *word) const {
  if (check_notrim(word)) return true;
  else return lang_->trim_n_try(*this, word);
}

const string aspell::to_soundslike(const string &word) const {
  return lang_->to_soundslike(word);
}

const aspell::SWordList aspell::soundslike(const string &word) const {
  return words_w_soundslike(lang_->to_soundslike(word));
}

const string aspell::to_phoneme(const string &word) const {
  return lang_->to_phoneme(word);
}

string ae_already_have_type::message() const {
  return "Sorry, I already have that type";
}

string ae_word_list_not_compellable::message() const {
  return "The current word list is not compatible with the requested id";
}
