// Aspell test program
// Copyright 1998 by Kevin Atkinson under the terms of the GPL

#include <string>
#include <rope>
#include <algo.h>
#include <iterator.h>
#include <deque>
#include <hash_map>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <strstream.h>
#include <time.h>

#include <unistd.h>

#include "as_config.h"

#include "popt.h"
#include "aspell.hh"
#include "check-t.hh"
#include "convert.hh"
#include "data.hh"
#include "const_string.hh"
#include "hash_string.hh"
#include "config.hh"

// action functions declarations

void pipe();
void print_ver();
void check();
void master();
void personal();
void repl();
void suggest();
void print_help();
void config();


/////////////////////////////////////////////////////////
//
// Command line options functions and classes
// (including main)
//

typedef deque<string> Args;
typedef StringMap Options;
enum action_t {do_create, do_merge, do_dump, do_other};

void options_act(poptContext, poptCallbackReason, 
		 const poptOption *, const char *, void *);

#define popt_string (POPT_ARG_STRING | POPT_ARGFLAG_ONEDASH)
#define popt_none   (POPT_ARG_NONE   | POPT_ARGFLAG_ONEDASH)

Args     args;
Options  options;
action_t action;
void (*fun) () = NULL;

#define H(name) (AspellConfig::help(#name))
#define ispell_comp(opt) \
 {NULL,opt,POPT_ARG_NONE,NULL,'i',"ispell compatibility option, ignored", NULL}
#define ispell_comp2(opt) \
 {NULL,opt,POPT_ARG_STRING,NULL,'i', \
  "ispell compatibility option, ignored", NULL}

poptOption popt_opts[] = { 
  {NULL,         '\0', POPT_ARG_CALLBACK, options_act, 0, NULL, NULL},
  {"master",          'd',  popt_string,  NULL, 'o', H(master),  "<base>"},
  {"personal",        'p',  popt_string,  NULL, 'o', H(personal),"<file>"},
  {"repl",            '\0', popt_string,  NULL, 'o', H(repl),    "<file>"},
  {"lang",            '\0', popt_string,  NULL, 'o', H(lang),    "<lang>"},
  {"dict-dir",        '\0', popt_string,  NULL, 'o', H(dict-dir), "<dir>"},
  {"data-dir",        '\0', popt_string,  NULL, 'o', H(data-dir), "<dir>"},
  {"ignore-repl",     '\0', popt_none ,   NULL, 'o', H(ignore-repl), NULL},
  {"dont-ignore-repl",'\0', popt_none,    NULL, 'o', NULL,           NULL},
  {"reverse",         '\0', popt_none,    NULL, 'o', 
     "reverse the order of the suggestion list", NULL   },
  {"dont-reverse",    '\0', popt_none,    NULL, 'o', NULL,           NULL},
  {"log",             '\0', popt_string,  NULL, 'o', 
     "output debug information to a log file",  "<file>"},
  {"conf",            '\0', popt_string,  NULL, 'o', H(conf),    "<file>"},
  {"conf-dir",        '\0', popt_string,  NULL, 'o', H(conf-dir), "<dir>"},
  {"create",          '\0', popt_string,  NULL, 'c', NULL,       "<what>"},
  {"dump",            '\0', popt_string,  NULL, 'c', NULL,       "<what>"},
  {"merge",           '\0', popt_string,  NULL, 'c', NULL,       "<what>"},
  {"check",           'c' , popt_none,    NULL, 'c', NULL,           NULL},
  {"pipe",            'a' , popt_none ,   NULL, 'c', NULL,           NULL},
  {"version",         'v' , popt_none ,   NULL, 'c', NULL,           NULL},
  {"help"   ,         '?' , popt_none ,   NULL, 'c', NULL,           NULL},
  ispell_comp('t'), ispell_comp('n'), ispell_comp('b'), ispell_comp('x'),
  ispell_comp('B'), ispell_comp('C'), ispell_comp('P'), ispell_comp('m'),
  ispell_comp('S'), ispell_comp2('w'), ispell_comp2('W'), ispell_comp2('T'),
  {0,0,0,0,0,0,0}
};

int main (int argc, const char *argv[]) 
try {
  if (argc == 1) {print_help(); return -1;}

  poptContext con = poptGetContext("aspell", argc, argv, popt_opts, 0);
  int rc = poptGetNextOpt(con);

  // 
  //
  
  if (!fun && rc == -1 && poptPeekArg(con)) {
    string arg = poptPeekArg(con);
    const char * argv[3] = {0,0,0};
    for(poptOption *i = popt_opts; i->argInfo; ++i) {
      if (i->val == 'c' && arg == i->longName) {
	poptGetArg(con);
	arg = "--" + arg;
	argv[0] = arg.c_str();
	if (i->argInfo == popt_string) argv[1] = poptGetArg(con);
	poptStuffArgs(con, argv);
	rc = poptGetNextOpt(con);
	break;
      }
    }
  } 
  
  if (rc != -1) {
    cerr << "Error: " << poptStrerror(rc) << ": " 
	 << poptBadOption(con, POPT_BADOPTION_NOALIAS) << endl;
    return -1;
  }

  const char * str;
  while ((str = poptGetArg(con))) {
    cout << str << "--" << endl;
    args.push_back(str);
  }

  poptFreeContext(con);

  if (options.have("log")) {
    ofstream LOG(options["log"].c_str(), ios::app);
    LOG << "Command: ";
    for(int i = 0; i != argc; ++i)
      LOG << argv[i] << "~";
    LOG << endl;
  }

  if (!fun) {
    cerr << "Error: You must specify an action\n";
    return -1;
  }
    
  (*fun)();

  return 0;

} catch (aspell_error& e) {

  cout << e.message() << endl;
  return -1;

}

void options_act(poptContext con, poptCallbackReason, 
		 const struct poptOption * opt,
		 const char * arg, void *) 
{
  if (!arg) arg = "";
  switch (opt->val) {

  case 'o': // the arg is an option

    options.replace(opt->longName, arg);
    break;

  case 'c': // the arg is an action to perform

    if (fun) {

      cerr << "Error: only one action may be specified\n";
      exit(-1);

    } else {

      string s = opt->longName;
      if      (s == "check")   fun = check;
      else if (s == "pipe")    fun = pipe;
      else if (s == "version") fun = print_ver;
      else if (s == "help")    fun = print_help;
      else if (s == "dump")    action = do_dump;
      else if (s == "merge")   action = do_merge;
      else if (s == "create")  action = do_create;
      else                     action = do_other;

      s = arg;
      if      (s == "master")   fun = master;
      else if (s == "personal") fun = personal;
      else if (s == "repl")     fun = repl;
      else if (s == "config")   fun = config;
      else if (s != "") {       
	cerr << "Error: unknown what for action " << opt->longName << endl;
	exit(-1);
      }
    }
    break;
  }
}


/////////////////////////////////////////////////////////
//
// Utility functions and classes
//

class bad_cin {};

class Ccin {
public:
  char get() {char c = cin.get(); if (!cin) throw bad_cin(); return c;}
  void unget() {cin.unget();}
  bool eof() {return cin.eof();}
};

Ccin& operator >> (Ccin &in, string &str) {
  cin >> str;
  if (!cin) throw bad_cin();
  return in;
}

void getline(Ccin, string &str) {
  getline(cin, str);
  if (!cin) throw bad_cin();
}

void getline(Ccin, string &str, char d) {
  getline(cin, str, d);
  if (!cin) throw bad_cin();
}

static Ccin ccin;

struct PrintStatus {
  bool print_star;
  PrintStatus() : print_star (true) {}
  void operator() () const {if (print_star) cout << "*" << endl;}
};

string trim_space(const string &s) {
  string::size_type beg = s.find_first_not_of(" \t\n");
  string::size_type end = s.find_last_not_of(" \t\n");
  if (beg == string::npos)
    return string();
  else
    return s.substr(beg,end-beg+1);
}

void get_word_pair(string &w1, string &w2, char sep = ',') {
  getline(ccin, w1, sep);
  w1 = trim_space(w1);
  getline(ccin, w2);
  w2 = trim_space(w2);
}

void ignore_rest() {
  cin.ignore(32767,'\n');
}

template<class Con>
void print_container(const Con& con) {
  typename Con::const_iterator i   = con.begin();
  typename Con::const_iterator end = con.end();
  cout << con.size() << ": ";
  if (i != end) {
    for (;;) {
      cout << *i;
      ++i;
      if (i == end) break;
      cout << ", ";
    }
  }
  cout << endl;
}

/////////////////////////////////////////////////////////
//
// Action Functions
//
//

///////////////////////////
//
// print_ver
//

void print_ver () {
  cout << "@(#) International Ispell Version 3.1.20 (but really Aspell " 
       << VERSION << " alpha)" << endl;
}

///////////////////////////
//
// print_help
//

void print_help () {
  cout << "
Aspell " VERSION " alpha.  Copyright 1999 by Kevin Atkinson.

Usgae: aspell [options] <command> 

<command> is one of:
  help          display this help message
  check <file>  to check a file
  pipe|-a       \"ispell -a\" compatibility mode 
  dump config   dumps the current configuration to stdout
  version       prints a version line
  dump|create|merge master|personal|repl [word list]
    dumps, creates or merges a master, personal, or replasment word list.

[options] is any of the following:
  -d, --master=<base>   main word list base name
  -p, --personal=<file> personal word list file name
  --repl=<file>         replacements list file name
  --lang=<lang>         default language to use
  --dict-dir=<dir>      location of the main word list
  --data-dir=<dir>      location of language data files
  --ignore-repl         ignore commands to store replacement pairs
  --dont-ignore-repl
  --log=<file>
  --conf=<file>         main configuration file
  --conf-dir=<dir>      location of main configuration file

The following options will be ignored for compatabilty with ispell:
  -b -m -n -t -x -B -C -P -S -w=ARG -T=ARG -W=ARG

";
}

///////////////////////////
//
// config
//

void config () {
  AspellConfig conf(options);
  conf.write_to_stream(cout);
}

///////////////////////////
//
// pipe
//

void pipe () {
  ofstream     LOG;
  bool         keep_log = false;
  bool         terse_mode = true;
  if (options.have("log")) {
    keep_log = true;
    LOG.open(options.lookup("log").c_str(), ios::app);
  }
  aspell sc(options);
  aspell_check_state<const_string::iterator>      
    state = url_skip_iterator<const_string::iterator>();
  aspell::Suggestions suggestions;
  unsigned int s;
  const_string::iterator i,j;
  const_string::iterator end;
  string line;
  string word;
  string word2;
  int    ignore;
  PrintStatus print_status;

  print_ver();
       
  char c;
  try {
    // break out of the loop on EOF via exceptions
    for (;;) {
      c = ccin.get();
      ignore = 0;
      switch (c) {
      case '*':
	ccin >> word;
	sc.add_personal(word);
	ignore_rest();
	break;
      case '&':
	ccin >> word;
	sc.add_personal(sc.lang().to_lower(word));
	ignore_rest();
      break;
      case '@':
	ccin >> word;
	sc.add_session(word);
	ignore_rest();
	break;
      case '#':
	sc.save_all_wls();
	ignore_rest();
	break;
      case '+':
      case '-':
	ignore_rest();
	break;
      case '!':
	terse_mode = true;
	print_status.print_star = false;
	ignore_rest();
	break;
      case '%':
	terse_mode = false;
	print_status.print_star = true;
	ignore_rest();
	break;
      case '^':
	ignore = 1;
      case '$':
	if (ccin.get() == '$') {
	  switch(ccin.get()) {
	  case 'm':
	    ccin >> word;
	    if (word == "which")
	      cout << state.itr().name() << endl;
	    else if (word == "none")
	      state = no_skip_iterator<const_string::iterator>();
	    else if (word == "url") 
	      state = url_skip_iterator<const_string::iterator>();
	    break;
	  case 's':
	    get_word_pair(word,word2);
	    cout << sc.score(word.c_str(),word2.c_str()) << endl;
	    break;
	  case 'S':
	    switch(ccin.get()) {
	    case 'W':
	    case 'w':
	      ccin >> word;
	      cout << sc.to_soundslike(word) << endl;
	      ignore_rest();
	      break;
	    case 'L':
	    case 'l':
	      ccin >> word;
	      print_container(sc.soundslike(word));
	      ignore_rest();
	      break;
	    default:
	      ignore_rest();
	    }
	    break;
	  case 'P':
	    switch(ccin.get()) {
	    case 'W':
	    case 'w':
	      ccin >> word;
	      cout << sc.to_phoneme(word) << endl;
	      ignore_rest();
	      break;
	    default:
	      ignore_rest();
	    }
	    break;
	  case 'r':
	    switch(ccin.get()) {
	    case 'a':
	      get_word_pair(word,word2);
	      sc.store_repl(word,word2);
	      break;
	    case 'i':
	      switch(ccin.get()) {
	      case 'c':
		cout << sc.ignore_repls() << endl;
		break;
	      case '0':
		sc.ignore_repls(0);
		break;
	      case '1':
		sc.ignore_repls(1);
		break;
	      }
	      ignore_rest();
	      break;
	    default:
	      ignore_rest();
	    }
	    break;
	  case 'p':
	    switch (ccin.get()) {
	    case 'p':
	      print_container(sc.personal_wl());
	      break;
	    case 's':
	      print_container(sc.session_wl());
	      break;
	    }
	    ignore_rest();
	    break;
	  case 'l':
	    cout << sc.lang_name() << endl;
	    ignore_rest();
	    break;
	  default:
	    ignore_rest();
	  }
	  break;
	} else {
	  ccin.unget();
	  // continue on (no break)
	}
      default:
	ccin.unget();
	getline(ccin, line);
	i = line.begin() + ignore;
	end = line.end();
	state.restart(i, end);
	while (check(sc, state, print_status)) {
	  word.resize(0);
	  for(j = state.word_begin(); j != state.word_end(); ++j) 
	    word +=*j;
	  suggestions = sc.suggest(word);
	  if (suggestions.size()) {
	    cout << "& " << word 
		 << " " << suggestions.size() 
		 << " " << state.word_begin() - line.begin()
		 << ":";
	    if (!options.have("reverse")) {
	      for (s = 0; s != suggestions.size() - 1; ++s)
		cout << " " << suggestions[s] << ",";
	      cout << " " << suggestions[suggestions.size() - 1] << endl;
	    } else {
	      for (s = suggestions.size()-1; s != 0; --s)
		cout << " " << suggestions[s] << ",";
	      cout << " " << suggestions[0] << endl;
	    }
	  } else {
	    cout << "# " << word << " " << state.word_begin() - line.begin() << endl;
	  }
	  state.advance();
	}
	cout << endl;
      }
    }
  } catch (bad_cin){
    //
  }
}

///////////////////////////
//
// check
//

void check () {
  crope file;
  string file_name = args[0];
  cout << file_name;
  {
    ifstream FILE(file_name.c_str());
    char c;
    while ((c = FILE.get()) != EOF) {
      file += c;
    }
  }

  aspell sc(options);

  aspell_check_state<crope::const_iterator> 
    state = no_skip_iterator<crope::const_iterator>();
  crope::const_iterator                      i,k;
  crope::size_type                           ii;
  crope::const_iterator  line_begin, line_end;
  string word, new_word;
  string choice;
  aspell::Suggestions suggestions;
  aspell::Suggestions::size_type suggestions_mid, suggestions_size;
  aspell::Suggestions::size_type j;

  hash_map<string,string,hash_string<string>,equal_to<string> > replace_list;
  cout.setf(ios::left);

  i = file.begin();
  state.restart(i, file.end());
  while (check(sc, state)) {
    word.resize(0);
    for (k = state.word_begin(); k != state.word_end(); ++k)
      word += *k;

    //
    // check if it is in the replace list
    //

    if (replace_list.count(word)) {
      ii = state.word_begin() - file.begin();
      file.replace(state.word_begin() - file.begin(), 
		   state.word_end() - state.word_begin(), 
		   replace_list[word].c_str());
      i = file.begin() + ii;
      state.restart(i, file.end());
      continue;
    }

    //
    // print the line with the misspelled word highlighted;
    //

    for (line_begin = state.word_begin(); 
	 line_begin != file.begin() && *line_begin != '\n';
	 --line_begin);
    if (line_begin != file.begin())
      ++line_begin;
    for (line_end = state.word_end();
	 line_end != file.end() && *line_end != '\n';
	 ++line_end);
    for(i = line_begin; i != state.word_begin(); ++i)
      cout << *i;
    cout << '*' << word << '*';
    i = state.word_end();
    for(; i != line_end; ++i) 
      cout << *i;
    cout << endl;
    
    //
    // print the suggestions and menu choices
    //

    suggestions = sc.suggest(word);
    suggestions_size = suggestions.size();
    if (suggestions_size > 9)
      suggestions_size = 9;
    suggestions_mid = suggestions_size / 2;
    if (suggestions_size % 2) suggestions_mid++; // if odd
    for (j = 0; j != suggestions_mid; ++j) {
      cout << setw(1) << j+1 << ") " << setw(37) << suggestions[j].c_str();
      if (j + suggestions_mid != suggestions_size) 
	cout << setw(1) << j+suggestions_mid+1 << ") " << suggestions[j+suggestions_mid];
      cout << endl;
    }
    cout << setw(40) << "i) Ignore"
	 << "I) Ignore all" << endl;
    cout << setw(40) << "r) Replace" 
	 << "R) Replace all" << endl;
    cout << setw(40) << "a) Add"
	 << setw(40) << "x) Exit" << endl;
    cout << "? ";

    //
    // Handle the users choice
    //

    ccin >> choice;
    
    switch (choice[0]) {
    case 'X':
    case 'x':
      goto exit_loop;
    case 'i':
      i = state.word_end();
      break;
    case 'I':
      sc.add_session(word);
      i = state.word_end();
      break;
    case 'a':
      sc.add_personal(word);
      i = state.word_end();
      break;
    case 'R':
    case 'r':
      cout << "With: ";
      ccin >> new_word;
      if (new_word[0] >= '1' && new_word[0] < suggestions_size + '1')
	new_word = suggestions[new_word[0]-'1'].c_str();
      ii = state.word_begin() - file.begin();
      file.replace(state.word_begin() - file.begin(), state.word_end() - state.word_begin(), new_word.c_str());
      sc.store_repl(word,new_word.c_str());
      i = file.begin() + ii;
      if (choice[0] == 'R')
	replace_list[word] = new_word;
      break;
    default:
      if (choice[0] >= '1' && choice[0] < suggestions_size + '1') { 
	ii = state.word_begin() - file.begin();
	file.replace(state.word_begin() - file.begin(), state.word_end() - state.word_begin(), suggestions[choice[0]-'1'].c_str());
	sc.store_repl(word,suggestions[choice[0]-'1'].c_str());
	i = file.begin() + ii;
      } else {
	cout << "Sorry that is an invalid choice!" << endl;
	i = state.word_begin();
      }
    }
    state.restart(i, file.end());
  }
 exit_loop:
  {
    ofstream FILE((file_name+".checked").c_str());
    FILE << file;
  }
  sc.save_all_wls();
}

///////////////////////////
//
// master
//

void master () {
  if (args.size() != 0) {
    options.replace("master", args[0].c_str());
  }

  AspellConfig config(options);
    

  if (action == do_create) {
    
    create_master(&config, istream_iterator<string>(cin),
		  istream_iterator<string>());

  } else if (action == do_merge) {
 
    cerr << "Can't merge a master word list yet.  Sorry\n";
    exit (-1);
  
  } else if (action == do_dump) {

    StandardWL_Base * mas = new ReadOnlyWL;
    mas->load(config.lookup("master-path"), &config);
    StandardWL_Base::iterator i   = mas->begin();
    StandardWL_Base::iterator end = mas->end();
    for (; i != end; ++i)
      cout << *i << endl;
  }
}

///////////////////////////
//
// personal
//

void personal () {
  if (args.size() != 0) {
    options.replace("personal", args[0].c_str());
  }
  if (action == do_create || action == do_merge) {
    aspell sc(options);

    if (action == do_create) {
      if (access(sc.config().lookup("personal-path").c_str(), F_OK) == 0) {
        cerr << "Sorry I won't overwrite \"" 
             << sc.config().lookup("personal-path") << "\"" << endl;
        exit (-1);
      }
      sc.personal_wl().clear();
    }

    string word;
    while (cin >> word) 
      sc.add_personal(word);

    sc.personal_wl().save();

  } else { // action == do_dump

    AspellConfig config(options);

    WritableWL per;
    per.load(config["personal-path"], &config);
    WritableWL::iterator i   = per.begin();
    WritableWL::iterator end = per.end();
    for (; i != end; ++i) {
      cout << *i << endl;
    }
    
  }
}

///////////////////////////
//
// repl
//

void repl() {
  if (args.size() != 0) {
    options.replace("repl", args[0].c_str());
  }

  if (action == do_create || action == do_merge) {

    aspell sc(options);

    if (action == do_create) {
      if (access(sc.config().lookup("repl-path").c_str(), F_OK) == 0) {
        cerr << "Sorry I won't overwrite \"" 
             << sc.config().lookup("repl-path") << "\"" << endl;
        exit (-1);
      }
      sc.personal_repl().clear();
    }
    
    try {
      string word,repl;

      while (true) {
	get_word_pair(word,repl,':');
	sc.store_repl(word,repl,false);
      }

    } catch (bad_cin) {}

    sc.personal_repl().save();
  
  } else if (action == do_dump) {

    AspellConfig config(options);

    WritableReplList repl;
    repl.load(config["repl-path"], &config);
    WritableReplList::iterator i   = repl.begin();
    WritableReplList::iterator end = repl.end();
    for (; i != end; ++i) {
      WritableReplList::Replacements::iterator j = (*i).begin();
      for (; j !=(*i).end() ; ++j) {
	cout << (*i).misspelled_word() << ": " << *j << endl;
      }
    }
  }
}
